/* This file is in the public domain */
#include <iostream>
#include <fstream>
#include <cctype>
#include <string>
#include <vector>

#include <opencl/pipe.h>
#include <opencl/hex.h>
#include <opencl/pk_types.h>

#include <opencl/randpool.h>
#include <opencl/rsa.h>
#include <opencl/dsa.h>

using namespace OpenCL;

SecureVector<byte> decode_hex(const std::string&);
BigInt to_bigint(const std::string&);

#define DEBUG 0

class Fixed_Output_RNG : public RandomNumberGenerator
   {
   public:
      byte random()
         {
         if(position < output.size())
            return output[position++];
         else
            return 42;
         }
      void clear() throw() {}
      std::string name() const { return "Fixed Output RNG"; }
      void add_randomness(const byte[], u32bit) throw() {}
      Fixed_Output_RNG(const SecureVector<byte>& x)
         {
         output = x;
         position = 0;
         }
   private:
      SecureVector<byte> output;
      u32bit position;
   };

std::vector<std::string> parse(const std::string&);
void strip(std::string&);

u32bit validate_dsa_sig(const std::string&, const std::vector<std::string>&);
u32bit validate_rsa_enc(const std::string&, const std::vector<std::string>&);
u32bit validate_rsa_sig(const std::string&, const std::vector<std::string>&);
u32bit validate_rw_sig(const std::string&, const std::vector<std::string>&);

int do_pk_validation_tests(const std::string& filename)
   {
   std::ifstream test_data(filename.c_str());

   if(!test_data)
       {
       std::cout << "Couldn't open test file " << filename << std::endl;
       std::exit(1);
       }

   u32bit errors = 0, alg_count = 0;
   std::string algorithm;

   while(!test_data.eof())
      {
      if(test_data.bad() || test_data.fail())
         {
         std::cout << "File I/O error." << std::endl;
         std::exit(1);
         }

      std::string line;
      std::getline(test_data, line);

      strip(line);
      if(line.size() == 0) continue;

      // Do line continuation
      while(line[line.size()-1] == '\\' && !test_data.eof())
         {
         line.replace(line.size()-1, 1, "");
         std::string nextline;
         std::getline(test_data, nextline);
         strip(nextline);
         if(nextline.size() == 0) continue;
         line += nextline;
         }

      if(line[0] == '[' && line[line.size() - 1] == ']')
         {
         algorithm = line.substr(1, line.size() - 2);
#if DEBUG
         std::cout << "Testing " << algorithm << std::endl;
#endif
         alg_count = 0;
         continue;
         }

      std::vector<std::string> substr = parse(line);

#if DEBUG
      std::cout << "Testing: " << algorithm << std::endl;
#endif

      u32bit new_errors = 0;
      if(algorithm.find("DSA") != std::string::npos)
         new_errors = validate_dsa_sig(algorithm, substr);

      if(algorithm.find("RSAES_") != std::string::npos)
         new_errors = validate_rsa_enc(algorithm, substr);

      if(algorithm.find("RSASSA_") != std::string::npos)
         new_errors = validate_rsa_sig(algorithm, substr);

      if(algorithm.find("RWSSA_") != std::string::npos)
         new_errors = validate_rw_sig(algorithm, substr);
      alg_count++;
      errors += new_errors;

      if(new_errors)
         std::cout << "ERROR: \"" << algorithm << "\" failed test #"
                   << std::dec << alg_count << std::endl;
      }

   // reset the global rng, because the validate_xxx will have replaced it
   set_global_rng(new Randpool);

   return errors;
   }

void dump_data(const SecureVector<byte>& out,
               const SecureVector<byte>& expected)
   {
   // For whatever reason, I cannot figure out how to do this with cout...
   std::printf("Got: ");
   for(u32bit j = 0; j != out.size(); j++)
      std::printf("%02X", out[j]);

   std::printf("\nExp: ");
   for(u32bit j = 0; j != expected.size(); j++)
      std::printf("%02X", expected[j]);
   std::printf("\n");
   std::fflush(stdout);
   }

void validate_encryption(PK_Encryptor* e, PK_Decryptor* d,
                         const std::string& algo, const std::string& input,
                         const std::string& random, const std::string& exp,
                         bool& failure)
   {
   SecureVector<byte> message = decode_hex(input);
   set_global_rng(new Fixed_Output_RNG(decode_hex(random)));

   SecureVector<byte> expected = decode_hex(exp);

   SecureVector<byte> out = e->encrypt(message);
   if(out != expected)
      {
      std::cout << "FAILED (encrypt): " << algo << std::endl;
      dump_data(out, expected);
      failure = true;
      }

   SecureVector<byte> decrypted = d->decrypt(out);
   if(decrypted != message)
      {
      std::cout << "FAILED (decrypt): " << algo << std::endl;
      failure = true;
      }

   delete e;
   delete d;
   }

void validate_signature(PK_Verifier* v, PK_Signer* s, const std::string& algo,
                        const std::string& input, const std::string& random,
                        const std::string& exp, bool& failure)
   {
   SecureVector<byte> message = decode_hex(input);
   set_global_rng(new Fixed_Output_RNG(decode_hex(random)));

   SecureVector<byte> expected = decode_hex(exp);

   SecureVector<byte> sig = s->sign_message(message, message.size());

   if(sig != expected)
      {
      std::cout << "FAILED (sign): " << algo << std::endl;
      dump_data(sig, expected);
      failure = true;
      }

   if(!v->verify_message(message, message.size(), sig, sig.size()))
      {
      std::cout << "FAILED (verify): " << algo << std::endl;
      failure = true;
      }

   /* This isn't a very thorough testing method, but it will hopefully
      catch any really horrible errors */
   sig[0]++;
   if(v->verify_message(message, message.size(), sig, sig.size()))
      {
      std::cout << "FAILED (accepted bad sig): " << algo << std::endl;
      failure = true;
      }

   delete v;
   delete s;
   }

u32bit validate_rsa_enc(const std::string& algo,
                        const std::vector<std::string>& str)
   {
   RSA_PrivateKey key(to_bigint(str[1]), to_bigint(str[2]),
                      to_bigint(str[0]));

   PK_Encryptor* e = 0;
   PK_Decryptor* d = 0;
   if(algo == "RSAES_EME1_SHA1")
      {
      e = new RSA_EME1_SHA1_Encryptor(key);
      d = new RSA_EME1_SHA1_Decryptor(key);
      }
   else if(algo == "RSAES_PKCS1v15")
      {
      e = new RSA_PKCS1v15_Encryptor(key);
      d = new RSA_PKCS1v15_Decryptor(key);
      }
   else if(algo == "RSAES_RAW")
      {
      e = new RSA_Raw_Encryptor(key);
      d = new RSA_Raw_Decryptor(key);
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_encryption(e, d, algo, str[3], str[4], str[5], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_rsa_sig(const std::string& algo,
                        const std::vector<std::string>& str)
   {
   if(str.size() < 6)
      throw Exception("Invalid input from pk_valid.dat");

   RSA_PrivateKey key(to_bigint(str[1]), to_bigint(str[2]),
                      to_bigint(str[0]));

   PK_Verifier* v = 0;
   PK_Signer* s = 0;
   if(algo == "RSASSA_EMSA2_SHA1")
      {
      v = new RSA_EMSA2_SHA1_Verifier(key);
      s = new RSA_EMSA2_SHA1_Signer(key);
      }
   else if(algo == "RSASSA_EMSA3_SHA1")
      {
      v = new RSA_EMSA3_SHA1_Verifier(key);
      s = new RSA_EMSA3_SHA1_Signer(key);
      }
   else if(algo == "RSASSA_EMSA3_MD5")
      {
      v = new RSA_EMSA3_MD5_Verifier(key);
      s = new RSA_EMSA3_MD5_Signer(key);
      }
   else if(algo == "RSASSA_RAW_SHA1")
      {
      v = new RSA_Raw_SHA1_Verifier(key);
      s = new RSA_Raw_SHA1_Signer(key);
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_signature(v, s, algo, str[3], str[4], str[5], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_rw_sig(const std::string& algo,
                       const std::vector<std::string>& str)
   {
   if(str.size() < 6)
      throw Exception("Invalid input from pk_valid.dat");

   RW_PrivateKey key(to_bigint(str[1]), to_bigint(str[2]), to_bigint(str[0]));

   PK_Verifier* v = 0;
   PK_Signer* s = 0;
   if(algo == "RWSSA_EMSA2_SHA1")
      {
      v = new RW_EMSA2_SHA1_Verifier(key);
      s = new RW_EMSA2_SHA1_Signer(key);
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_signature(v, s, algo, str[3], str[4], str[5], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_dsa_sig(const std::string& algo,
                        const std::vector<std::string>& str)
   {
   if(str.size() != 8)
      throw Exception("Invalid input from pk_valid.dat");

   DSA_PrivateKey key(to_bigint(str[0]), to_bigint(str[1]),
                      to_bigint(str[2]), to_bigint(str[3]),
                      to_bigint(str[4]));

   PK_Verifier* v = new DSA_Verifier(key);
   PK_Signer* s = new DSA_Signer(key);

   bool failure = false;
   validate_signature(v, s, algo, str[5], str[6], str[7], failure);
   return (failure ? 1 : 0);
   }
