#include <opencl/pk_types.h>
using namespace OpenCL;

#include <iostream>
#include <fstream>

#define DEBUG 0

BigInt to_bigint(const std::string&);

DSA_PrivateKey load_dsa_key(const std::string&);
RSA_PrivateKey load_rsa_key(const std::string&);
RW_PrivateKey  load_rw_key(const std::string&);

void bench_enc(PK_Encryptor*, const std::string&, double, bool);
void bench_dec(PK_Encryptor*, PK_Decryptor*, const std::string&, double, bool);
void bench_sig(PK_Signer*, const std::string&, double, bool);
void bench_ver(PK_Signer*, PK_Verifier*, const std::string&, double, bool);

void bench_rsa_set(RSA_PrivateKey& key, const std::string keybits,
                   double seconds, bool html)
   {
   bench_enc(new RSA_Raw_Encryptor(key),
             "RSA-" + keybits, seconds, html);
   bench_dec(new RSA_Raw_Encryptor(key), new RSA_Raw_Decryptor(key),
             "RSA-" + keybits, seconds, html);
   }

void bench_rw_set(RW_PrivateKey& key, const std::string keybits,
                  double seconds, bool html)
   {
   bench_ver(new RW_EMSA2_SHA1_Signer(key), new RW_EMSA2_SHA1_Verifier(key),
             "RW-" + keybits, seconds, html);
   bench_sig(new RW_EMSA2_SHA1_Signer(key),
             "RW-" + keybits, seconds, html);
   }

void bench_dsa_set(DSA_PrivateKey& key, const std::string keybits,
                   double seconds, bool html)
   {
   bench_ver(new DSA_Signer(key), new DSA_Verifier(key),
             "DSA-" + keybits, seconds, html);
   bench_sig(new DSA_Signer(key),
             "DSA-" + keybits, seconds, html);
   }

void bench_pk(bool html, double seconds)
   {
   try {

   DSA_PrivateKey dsa1024 = load_dsa_key("checks/keys/dsa1024.key");
   bench_dsa_set(dsa1024, "1024", seconds, html);

   RSA_PrivateKey rsa512 = load_rsa_key("checks/keys/rsa512.key");
   RSA_PrivateKey rsa1024 = load_rsa_key("checks/keys/rsa1024.key");
   RSA_PrivateKey rsa2048 = load_rsa_key("checks/keys/rsa2048.key");

   bench_rsa_set(rsa512, "512", seconds, html);
   bench_rsa_set(rsa1024, "1024", seconds, html);
   bench_rsa_set(rsa2048, "2048", seconds, html);

   RW_PrivateKey rw512 = load_rw_key("checks/keys/rw512.key");
   RW_PrivateKey rw1024 = load_rw_key("checks/keys/rw1024.key");
   bench_rw_set(rw512, "512", seconds, html);
   bench_rw_set(rw1024, "1024", seconds, html);

   }
   catch(OpenCL::Exception& e)
      {
      std::cout << "Exception caught: " << e.what() << std::endl;
      return;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: "
                << e.what() << std::endl;
      return;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      return;
      }

   }

void print_result(bool, u32bit runs, u32bit clocks_used,
                  const std::string& algo_name, const std::string& op)
   {
   std::cout << algo_name << ": ";

   double seconds = (double)clocks_used / CLOCKS_PER_SEC;
   double mseconds_per_run = 1000 * (seconds / runs);

   std::cout.setf(std::ios::fixed, std::ios::floatfield);
   std::cout.precision(2);
   std::cout << mseconds_per_run << " ms / " << op << std::endl;
   }

void bench_enc(PK_Encryptor* enc, const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(msg, MSG_SIZE);
      enc->encrypt(msg, MSG_SIZE);
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "public operation");
   }

void bench_dec(PK_Encryptor* enc, PK_Decryptor* dec,
               const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];
   Global_RNG::randomize(msg, MSG_SIZE);

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   SecureVector<byte> encrypted_msg = enc->encrypt(msg, MSG_SIZE);

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      dec->decrypt(encrypted_msg, encrypted_msg.size());
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "private operation");
   }

void bench_sig(PK_Signer* sig, const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(msg, MSG_SIZE);
      sig->update(msg, MSG_SIZE);
      sig->signature();
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "private operation");
   }

void bench_ver(PK_Signer* sig, PK_Verifier* ver,
               const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];
   Global_RNG::randomize(msg, MSG_SIZE);

   sig->update(msg, MSG_SIZE);
   SecureVector<byte> signature = sig->signature();

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      ver->update(msg, MSG_SIZE);
      ver->valid_signature(signature, signature.size());
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "public operation");
   }

RSA_PrivateKey load_rsa_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the RSA key file " + file);

   std::string e, p, q;

   std::getline(keyfile, e);
   std::getline(keyfile, p);
   std::getline(keyfile, q);

   RSA_PrivateKey key(to_bigint(p), to_bigint(q), to_bigint(e));
#if DEBUG
   key.check_params();
#endif
   return key;
   }

RW_PrivateKey load_rw_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the RW key file " + file);

   std::string e, p, q;

   std::getline(keyfile, e);
   std::getline(keyfile, p);
   std::getline(keyfile, q);

   RW_PrivateKey key(to_bigint(p), to_bigint(q), to_bigint(e));
#if DEBUG
   key.check_params();
#endif
   return key;
   }

DSA_PrivateKey load_dsa_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the DSA key file " + file);

   std::string p, q, g, y;

   std::getline(keyfile, p);
   std::getline(keyfile, q);
   std::getline(keyfile, g);
   std::getline(keyfile, y);

   DSA_PrivateKey key(to_bigint(p), to_bigint(q),
                      to_bigint(g), to_bigint(y));
#if DEBUG
   key.check_params();
#endif
   return key;
   }
