/*************************************************
* DSA Source File                                *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/dsa.h>
#include <opencl/numthry.h>
#include <opencl/primes.h>

#include <assert.h>

namespace OpenCL {

/*************************************************
* DSA_PublicKey Constructor                      *
*************************************************/
DSA_PublicKey::DSA_PublicKey(const BigInt& p1, const BigInt& q1,
                             const BigInt& g1, const BigInt& y1) :
   p(p1), q(q1), g(g1), y(y1), powermod_g_p(g, p), powermod_y_p(y, p)
   {
   reducer_q = get_reducer(q);
   }

/*************************************************
* DSA_PublicKey Copy Constructor                 *
*************************************************/
DSA_PublicKey::DSA_PublicKey(const DSA_PublicKey& key) :
   PK_Verifying_wo_MR_Key(), p(key.p), q(key.q), g(key.g), y(key.y),
   powermod_g_p(g, p), powermod_y_p(y, p)
   {
   reducer_q = get_reducer(q);
   }

/*************************************************
* DSA_PublicKey Destructor                       *
*************************************************/
DSA_PublicKey::~DSA_PublicKey()
   {
   delete reducer_q;
   }

/*************************************************
* DSA_PublicKey Assignment                       *
*************************************************/
DSA_PublicKey& DSA_PublicKey::operator=(const DSA_PublicKey& key)
   {
   p = key.p;
   q = key.q;
   g = key.g;
   y = key.y;
   reducer_q = get_reducer(q);
   powermod_g_p = key.powermod_g_p;
   powermod_y_p = key.powermod_y_p;
   return (*this);
   }

/*************************************************
* Check the public DSA parameters for validity   *
*************************************************/
bool DSA_PublicKey::check_params() const
   {
   if((p - BigInt::one()) % q != 0)
      return false;
   if(g == BigInt::zero() || g == BigInt::one() || g >= p)
      return false;
   if(!is_prime(p) || !is_prime(q))
      return false;
   return true;
   }

/*************************************************
* DSA Verification Function                      *
*************************************************/
bool DSA_PublicKey::verify(const byte msg[], u32bit msg_len,
                           const byte sig[], u32bit sig_len) const
   {
   if(sig_len % 2 == 1)
      return false;

   t1.binary_decode(msg, msg_len);
   r.binary_decode(sig, sig_len / 2);
   s.binary_decode(sig + sig_len / 2, sig_len / 2);

   if(r.is_zero() || s.is_zero() || r >= q || s >= q)
      return false;

   BigInt w = inverse_mod(s, q);
   BigInt u1 = reducer_q->multiply(w, t1);
   BigInt u2 = reducer_q->multiply(w, r);
   BigInt v = (powermod_g_p(u1) * powermod_y_p(u2)) % p;
   return (reducer_q->reduce(v) == r);
   }

/*************************************************
* Create a DSA private key                       *
*************************************************/
DSA_PrivateKey::DSA_PrivateKey(const BigInt& p1, const BigInt& q1,
                               const BigInt& g1, const BigInt& y1,
                               const BigInt& x1) :
   DSA_PublicKey(p1, q1, g1, y1), x(x1)
   {
   if(!x) x = random_integer(q.bits() - 1);
   if(!y) y = powermod_g_p(x);
   }

/*************************************************
* Create a new DSA key                           *
*************************************************/
DSA_PrivateKey::DSA_PrivateKey(u32bit pbits, u32bit qbits,
                               bool use_fips_keygen)
   {
   if(pbits - 16 <= qbits)
      throw Invalid_Argument("DSA_PrivateKey: Invalid sizes for primes");

   if(use_fips_keygen && (!valid_fips186_p_size(pbits) || (qbits != 160)))
      throw Invalid_Argument("DSA_PrivateKey: Primes sizes not legal for "
                             "FIPS-186 prime generation");

   if(use_fips_keygen)
      generate_dsa_primes(p, q, pbits);
   else
      {
      q = random_prime(qbits);
      do
         p = random_integer(pbits - qbits) * q + 1;
      while(!is_prime(p) || p.bits() != pbits);
      }

   keygen_operation();
   }

/*************************************************
* Create a private DSA key from a seed           *
*************************************************/
DSA_PrivateKey::DSA_PrivateKey(const byte seed[], u32bit seed_length,
                               u32bit pbits)
   {
   if(!generate_dsa_primes(p, q, seed, seed_length, pbits))
      throw Invalid_Argument("DSA_PrivateKey: The seed value given does not "
                             "generate a DSA key");
   keygen_operation();
   }

/*************************************************
* Handle Parts of DSA Key Generation             *
*************************************************/
void DSA_PrivateKey::keygen_operation()
   {
   BigInt e = (p - 1) / q;
   u32bit prime_no = 0;
   do
      g = power_mod(PRIMES[prime_no++], e, p);
   while(g == BigInt::one());

   x = random_integer(q.bits() - 1);

   powermod_g_p = FixedBase_Exp(g, p);
   reducer_q = get_reducer(q);

   y = powermod_g_p(x);
   powermod_y_p = FixedBase_Exp(y, p);
   }

/*************************************************
* Check the private DSA Parameters for validity  *
*************************************************/
bool DSA_PrivateKey::check_params() const
   {
   if(!DSA_PublicKey::check_params())
      return false;
   if(y != powermod_g_p(x))
      return false;
   return true;
   }

/*************************************************
* Check for a p size authorized by FIPS 186-2    *
*************************************************/
bool DSA_PrivateKey::valid_fips186_p_size(u32bit pbits)
   {
   return ((pbits % 64 == 0) || (pbits >= 512) || (pbits <= 1024));
   }

/*************************************************
* DSA Signature Operation                        *
*************************************************/
SecureVector<byte> DSA_PrivateKey::sign(const byte in[], u32bit len) const
   {
   t1.binary_decode(in, len);

   k = random_integer(q.bits());
   while(k >= q)
      k = random_integer(q.bits());

   r = reducer_q->reduce(powermod_g_p(k));
   s = reducer_q->multiply(inverse_mod(k, q), mul_add(x, r, t1));

   SecureVector<byte> output(2*q.bytes());
   r.binary_encode(output + (q.bytes() - r.bytes()));
   s.binary_encode(output + output.size() / 2 + (q.bytes() - s.bytes()));
   return output;
   }

}
