/*************************************************
* Public Key Base Source File                    *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/pubkey.h>
#include <opencl/util.h>

namespace OpenCL {

/*************************************************
* PK_Encryptor Constructor                       *
*************************************************/
PK_Encryptor::PK_Encryptor(const PK_Encrypting_Key& k, EME* me) :
   key(k), encoder(me)
   {
   }

/*************************************************
* Encrypt a message                              *
*************************************************/
SecureVector<byte> PK_Encryptor::encrypt(const byte msg[],
                                         u32bit msg_len) const
   {
   SecureVector<byte> message;
   if(encoder) message = encoder->encode(msg, msg_len, key.max_input_bits());
   else        message.set_to(msg, msg_len);

   if(8*(message.size() - 1) + high_bit(message[0]) > key.max_input_bits())
      throw Exception("PK_Encryptor: Input is too large");

   return key.encrypt(message, message.size());
   }

/*************************************************
* Encrypt a message                              *
*************************************************/
SecureVector<byte> PK_Encryptor::encrypt(const SecureVector<byte>& msg) const
   {
   return encrypt(msg, msg.size());
   }

/*************************************************
* Return the max size, in bytes, of a message    *
*************************************************/
u32bit PK_Encryptor::maximum_input_size() const
   {
   if(!encoder)
      return (key.max_input_bits() / 8);
   else
      return encoder->maximum_input_size(key.max_input_bits());
   }

/*************************************************
* PK_Decryptor Constructor                       *
*************************************************/
PK_Decryptor::PK_Decryptor(const PK_Decrypting_Key& k, EME* me) :
   key(k), encoder(me)
   {
   }

/*************************************************
* Decrypt a message                              *
*************************************************/
SecureVector<byte> PK_Decryptor::decrypt(const byte in[], u32bit length) const
   {
   try {
      SecureVector<byte> decrypted = key.decrypt(in, length);
      if(encoder)
         return encoder->decode(decrypted, key.max_input_bits());
      else
         return decrypted;
      }
   catch(Invalid_Argument)
      {
      throw Exception("PK_Decryptor: Input is invalid");
      }
   catch(Decoding_Error)
      {
      throw Exception("PK_Decryptor: Input is invalid");
      }
   }

/*************************************************
* Decrypt a message                              *
*************************************************/
SecureVector<byte> PK_Decryptor::decrypt(const SecureVector<byte>& msg) const
   {
   return decrypt(msg, msg.size());
   }

/*************************************************
* PK_Signer Constructor                          *
*************************************************/
PK_Signer::PK_Signer(const PK_Signing_Key& k, HashFunction* h, EMSA* se) :
   key(k), hash(h), encoder(se)
   {
   }

/*************************************************
* Sign a message                                 *
*************************************************/
SecureVector<byte> PK_Signer::sign_message(const byte msg[], u32bit length)
   {
   update(msg, length);
   return signature();
   }

/*************************************************
* Update the hash, or append to the message      *
*************************************************/
void PK_Signer::update(const byte in[], u32bit length)
   {
   if(hash)
      hash->update(in, length);
   else
      message.append(in, length);
   }

/*************************************************
* Create a signature                             *
*************************************************/
SecureVector<byte> PK_Signer::signature()
   {
   if(hash)
      message = hash->final();
   if(encoder)
      message = encoder->encode(message, key.max_input_bits());

   SecureVector<byte> sig = key.sign(message, message.size());
   message.destroy();
   return sig;
   }

/*************************************************
* PK_Verifier Constructor                        *
*************************************************/
PK_Verifier::PK_Verifier(HashFunction* h, EMSA* se) : encoder(se), hash(h)
   {
   }

/*************************************************
* Verify a message                               *
*************************************************/
bool PK_Verifier::verify_message(const byte msg[], u32bit msg_length,
                                 const byte sig[], u32bit sig_length)
   {
   update(msg, msg_length);
   return valid_signature(sig, sig_length);
   }

/*************************************************
* Append to the message                          *
*************************************************/
void PK_Verifier::update(const byte in[], u32bit length)
   {
   if(hash)
      hash->update(in, length);
   else
      message.append(in, length);
   }

/*************************************************
* Create a signature                             *
*************************************************/
bool PK_Verifier::valid_signature(const byte sig[], u32bit length)
   {
   if(hash)
      message = hash->final();

   try {
      bool valid = validate_signature(message, sig, length);
      message.destroy();
      return valid;
   }
   catch(Invalid_Argument)
      {
      message.destroy();
      return false;
      }
   }

/*************************************************
* Verify a signature                             *
*************************************************/
bool PK_Verifier_with_MR::validate_signature(const SecureVector<byte>& msg,
                                             const byte sig[], u32bit sig_len)
   {
   SecureVector<byte> output_of_key = key.verify(sig, sig_len);
   if(encoder)
      return encoder->is_valid(output_of_key, msg, key.max_input_bits());
   else
      return (msg == output_of_key);
   }

/*************************************************
* PK_Verifier_with_MR Constructor                *
*************************************************/
PK_Verifier_with_MR::PK_Verifier_with_MR(const PK_Verifying_with_MR_Key& k,
                                         HashFunction* h, EMSA* se) :
   PK_Verifier(h, se), key(k)
   {
   }

/*************************************************
* Verify a signature                             *
*************************************************/
bool PK_Verifier_wo_MR::validate_signature(const SecureVector<byte>& msg,
                                           const byte sig[], u32bit sig_len)
   {
   return key.verify(msg, msg.size(), sig, sig_len);
   }

/*************************************************
* PK_Verifier_wo_MR Constructor                  *
*************************************************/
PK_Verifier_wo_MR::PK_Verifier_wo_MR(const PK_Verifying_wo_MR_Key& k,
                                     HashFunction* h) :
   PK_Verifier(h, 0), key(k)
   {
   }


}
